"""
Script for generating the model zoo docs page contents
``docs/source/user_guide/model_zoo/models.rst``.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import os
import re

from jinja2 import Environment, BaseLoader

import eta.core.utils as etau

import fiftyone.zoo as foz


logger = logging.getLogger(__name__)


_HEADER = """
.. _model-zoo-models:

Available Zoo Models
====================

.. default-role:: code

This page lists all of the models available in the Model Zoo.

.. note::

    Check out the :ref:`API reference <model-zoo-api>` for complete
    instructions for using the Model Zoo.
"""


_SECTION_TEMPLATE = """
.. _model-zoo-{{ link_name }}-models:

{{ header_name }} models
{{ '-' * (header_name|length + 7) }}
"""


_MODEL_TEMPLATE = """
.. _model-zoo-{{ name }}:

{{ header_name }}
{{ '_' * header_name|length }}

{{ description }}.

**Details**

-   Model name: ``{{ name }}``
-   Model source: {{ source }}
-   Model size: {{ size }}
-   Exposes embeddings? {{ exposes_embeddings }}
-   Tags: ``{{ tags }}``

**Requirements**

{% if base_packages %}
-   Packages: ``{{ base_packages }}``

{% endif %}
-   CPU support

    -   {{ supports_cpu }}
{% if cpu_packages %}
    -   Packages: ``{{ cpu_packages }}``
{% endif %}

-   GPU support

    -   {{ supports_gpu }}
{% if gpu_packages %}
    -   Packages: ``{{ gpu_packages }}``
{% endif %}

**Example usage**

.. code-block:: python
    :linenos:

    import fiftyone as fo
    import fiftyone.zoo as foz

{% if 'imagenet' in name %}
    dataset = foz.load_zoo_dataset(
        "imagenet-sample",
        dataset_name=fo.get_default_dataset_name(),
        max_samples=50,
        shuffle=True,
    )
{% else %}
    dataset = foz.load_zoo_dataset(
        "coco-2017",
        split="validation",
        dataset_name=fo.get_default_dataset_name(),
        max_samples=50,
        shuffle=True,
    )
{% endif %}

    model = foz.load_zoo_model("{{ name }}")

    dataset.apply_model(model, label_field="predictions")

    session = fo.launch_app(dataset)

{% if 'zero-shot' in tags %}
    #
    # Make zero-shot predictions with custom classes
    #

    model = foz.load_zoo_model(
        "{{ name }}",
        text_prompt="A photo of a",
        classes=["person", "dog", "cat", "bird", "car", "tree", "chair"],
    )

    dataset.apply_model(model, label_field="predictions")
    session.refresh()
{% endif %}
"""


_CARD_SECTION_START = """
.. raw:: html

    <div id="tutorial-cards-container">

    <nav class="navbar navbar-expand-lg navbar-light tutorials-nav col-12">
        <div class="tutorial-tags-container">
            <div id="dropdown-filter-tags">
                <div class="tutorial-filter-menu">
                    <div class="tutorial-filter filter-btn all-tag-selected" data-tag="all">All</div>
                </div>
            </div>
        </div>
    </nav>

    <hr class="tutorials-hr">

    <div class="row">

    <div id="tutorial-cards">
    <div class="list">
"""


_CARD_SECTION_END = """
.. raw:: html

    </div>

    <div class="pagination d-flex justify-content-center"></div>

    </div>

    </div>
"""


_CARD_MODEL_TEMPLATE = """
.. customcarditem::
    :header: {{ header }}
    :description: {{ description }}
    :link: {{ link }}
    :tags: {{ tags }}
"""


def _render_section_content(template, all_models, print_source, header_name):
    models = []
    for model_name, source, _ in all_models:
        if source != print_source:
            continue

        zoo_model = foz.get_zoo_model(model_name)

        tags_str = ", ".join(zoo_model.tags)

        models.append({"name": model_name, "tags_str": tags_str})

    col1_width = 2 * max(len(m["name"]) for m in models) + 22
    col2_width = max(len(m["tags_str"]) for m in models) + 2

    return template.render(
        link_name=print_source,
        header_name=header_name,
        col1_width=col1_width,
        col2_width=col2_width,
        models=models,
    )


def _render_model_content(template, model_name):
    zoo_model = foz.get_zoo_model(model_name)

    if "torch" in zoo_model.tags:
        source = "torch"
    elif any(t in zoo_model.tags for t in ("tf", "tf1", "tf2")):
        source = "tensorflow"
    else:
        source = "other"

    header_name = model_name

    size_str = etau.to_human_bytes_str(zoo_model.size_bytes, decimals=2)
    size_str = size_str[:-2] + " " + size_str[-2:]  # 123.45 MB, not 123.45MB

    if "embeddings" in zoo_model.tags:
        exposes_embeddings = "yes"
    else:
        exposes_embeddings = "no"

    tags_str = ", ".join(zoo_model.tags)

    base_packages = zoo_model.requirements.packages
    if base_packages is not None:
        base_packages = ", ".join(base_packages)

    if zoo_model.supports_cpu:
        supports_cpu = "yes"
    else:
        supports_cpu = "no"

    cpu_packages = zoo_model.requirements.cpu_packages
    if cpu_packages is not None:
        cpu_packages = ", ".join(cpu_packages)

    if zoo_model.supports_gpu:
        supports_gpu = "yes"
    else:
        supports_gpu = "no"

    gpu_packages = zoo_model.requirements.gpu_packages
    if gpu_packages is not None:
        gpu_packages = ", ".join(gpu_packages)

    content = template.render(
        name=zoo_model.name,
        header_name=header_name,
        description=zoo_model.description,
        source=zoo_model.source,
        size=size_str,
        exposes_embeddings=exposes_embeddings,
        tags=tags_str,
        base_packages=base_packages,
        supports_cpu=supports_cpu,
        cpu_packages=cpu_packages,
        supports_gpu=supports_gpu,
        gpu_packages=gpu_packages,
    )

    return source, content


def _render_card_model_content(template, model_name):
    zoo_model = foz.get_zoo_model(model_name)

    tags = []

    for tag in zoo_model.tags:
        if tag == "tf1":
            tags.append("TensorFlow-1")
        elif tag == "tf2":
            tags.append("TensorFlow-2")
        elif tag == "tf":
            tags.append("TensorFlow")
        elif tag == "torch":
            tags.append("PyTorch")
        else:
            tags.append(tag.capitalize().replace(" ", "-"))

    tags = ",".join(tags)

    link = "models.html#%s" % zoo_model.name

    description = zoo_model.description

    # remove paper links from descriptions
    description = description.replace("`_", '"')
    description = description.replace("`", '"')
    description = re.sub(" <.*>", "", description)

    content = template.render(
        header=zoo_model.name, description=description, link=link, tags=tags
    )

    return content


def _generate_section(template, all_models, print_source, header_name):
    content = [
        _render_section_content(
            template, all_models, print_source, header_name
        )
    ]

    for _, source, model_content in all_models:
        if source == print_source:
            content.append(model_content)

    return content


def main():
    # Render model sections

    environment = Environment(
        loader=BaseLoader,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    section_template = environment.from_string(_SECTION_TEMPLATE)
    model_template = environment.from_string(_MODEL_TEMPLATE)
    card_model_template = environment.from_string(_CARD_MODEL_TEMPLATE)

    models = []
    for model_name in foz.list_zoo_models():
        source, content = _render_model_content(model_template, model_name)
        models.append((model_name, source, content))

    # Generate page content

    content = [_HEADER]
    content.append(_CARD_SECTION_START)
    for model_name in foz.list_zoo_models():
        card_content = _render_card_model_content(
            card_model_template, model_name
        )
        content.append(card_content)

    content.append(_CARD_SECTION_END)
    content.extend(
        _generate_section(section_template, models, "torch", "Torch")
    )
    content.extend(
        _generate_section(section_template, models, "tensorflow", "TensorFlow")
    )

    # Write docs page

    docs_dir = "/".join(os.path.realpath(__file__).split("/")[:-2])
    outpath = os.path.join(docs_dir, "source/user_guide/model_zoo/models.rst")

    print("Writing '%s'" % outpath)
    etau.write_file("\n".join(content), outpath)


if __name__ == "__main__":
    main()



"""
Backing document classes for :class:`fiftyone.core.sample.Sample` instances.

Class hierarchy::

    SerializableDocument
    ├── NoDatasetSampleDocument
    └── DatasetSampleDocument
        ├── my_custom_dataset
        ├── another_dataset
        └── ...

Design invariants:

-   A :class:`fiftyone.core.sample.Sample` always has a backing ``_doc`` that
    is an instance of :class:`fiftyone.core.odm.document.SerializableDocument`

-   A :class:`fiftyone.core.dataset.Dataset` always has a backing
    `_sample_doc_cls` that is a subclass of :class:`DatasetSampleDocument`

**Implementation details**

When a new :class:`fiftyone.core.sample.Sample` is created, its ``_doc``
attribute is an instance of :class:`NoDatasetSampleDocument`::

    import fiftyone as fo

    sample = fo.Sample()
    sample._doc  # NoDatasetSampleDocument

When a new :class:`fiftyone.core.dataset.Dataset` is created, its
``_sample_doc_cls`` attribute holds a dynamically created subclass of
:class:`DatasetSampleDocument` whose name is the name of the dataset's sample
collection::

    dataset = fo.Dataset(name="my_dataset")
    dataset._sample_doc_cls  # my_dataset(DatasetSampleDocument)

When a sample is added to a dataset, its ``_doc`` attribute is changed from
type :class:`NoDatasetSampleDocument` to type ``dataset._sample_doc_cls``::

    dataset.add_sample(sample)
    sample._doc  # my_dataset(DatasetSampleDocument)

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import OrderedDict
import random

import fiftyone.core.fields as fof
import fiftyone.core.metadata as fom
import fiftyone.core.media as fomm
import fiftyone.core.utils as fou

from .document import Document, SerializableDocument
from .mixins import DatasetMixin, get_default_fields, NoDatasetMixin


# Use our own Random object to avoid messing with the user's seed
_random = random.Random()


def _generate_rand(filepath=None):
    # @todo filepath no longer has to be unique. Should we change this?
    if filepath is not None:
        _random.seed(filepath)

    return _random.random() * 0.001 + 0.999


class DatasetSampleDocument(DatasetMixin, Document):
    """Base class for sample documents backing samples in datasets.

    All ``fiftyone.core.dataset.Dataset._sample_doc_cls`` classes inherit from
    this class.
    """

    meta = {"abstract": True}

    _is_frames_doc = False

    id = fof.ObjectIdField(required=True, primary_key=True, db_field="_id")
    filepath = fof.StringField(required=True)
    tags = fof.ListField(fof.StringField())
    metadata = fof.EmbeddedDocumentField(fom.Metadata, null=True)

    _media_type = fof.StringField()
    _rand = fof.FloatField(default=_generate_rand)

    @property
    def media_type(self):
        return self._media_type

    def _get_repr_fields(self):
        fields = self.field_names
        return fields[:1] + ("media_type",) + fields[1:]


class NoDatasetSampleDocument(NoDatasetMixin, SerializableDocument):
    """Backing document for samples that have not been added to a dataset."""

    _is_frames_doc = False

    # pylint: disable=no-member
    default_fields = DatasetSampleDocument._fields
    default_fields_ordered = get_default_fields(
        DatasetSampleDocument, include_private=True
    )

    def __init__(self, **kwargs):
        filepath = fou.normalize_path(kwargs["filepath"])

        kwargs["id"] = kwargs.get("id", None)
        kwargs["filepath"] = filepath
        kwargs["_rand"] = _generate_rand(filepath=filepath)
        kwargs["_media_type"] = fomm.get_media_type(filepath)

        self._data = OrderedDict()

        for field_name in self.default_fields_ordered:
            value = kwargs.pop(field_name, None)

            if value is None and field_name != "id":
                value = self._get_default(self.default_fields[field_name])

            self._data[field_name] = value

        self._data.update(kwargs)

    @property
    def media_type(self):
        return self._media_type

    def _get_repr_fields(self):
        fields = self.field_names
        return fields[:1] + ("media_type",) + fields[1:]



"""
FiftyOne v0.14.0 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    if "created_at" not in dataset_dict:
        dataset_dict["created_at"] = None

    if "last_loaded_at" not in dataset_dict:
        dataset_dict["last_loaded_at"] = None

    if "frame_collection_name" not in dataset_dict:
        name = "frames." + dataset_dict["sample_collection_name"]
        dataset_dict["frame_collection_name"] = name

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    dataset_dict.pop("created_at", None)
    dataset_dict.pop("last_loaded_at", None)
    dataset_dict.pop("frame_collection_name", None)

    db.datasets.replace_one(match_d, dataset_dict)



"""
FiftyOne Server /events route

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import typing as t

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from sse_starlette.sse import EventSourceResponse

from fiftyone.server.events import (
    ListenPayload,
    add_event_listener,
    dispatch_polling_event_listener,
)
from fiftyone.server.decorators import route


class Events(HTTPEndpoint):
    @route
    async def post(
        self, request: Request, data: dict
    ) -> t.Union[t.Dict, EventSourceResponse]:
        polling = data.pop("polling", False)
        payload = ListenPayload.from_dict(data)
        if polling:
            return await dispatch_polling_event_listener(request, payload)

        return EventSourceResponse(
            add_event_listener(request, payload),
            ping=2,
        )



"""
Sample groups.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from copy import deepcopy

from bson import ObjectId

import fiftyone.core.fields as fof
import fiftyone.core.odm as foo


class Group(foo.EmbeddedDocument):
    """A named group membership.

    Args:
        id (None): the group ID
        name (None): the group name
    """

    id = fof.ObjectIdField(default=lambda: str(ObjectId()), db_field="_id")
    name = fof.StringField()

    def __deepcopy__(self, memo):
        # Custom implementation that does NOT exclude `id` field, since `Group`
        # instances often do need to share the same IDs
        # pylint: disable=no-member, unsubscriptable-object
        return self.__class__(
            **{
                f: deepcopy(self.get_field(f), memo)
                for f in self._fields_ordered
            }
        )

    @property
    def _id(self):
        return ObjectId(self.id)

    def element(self, name):
        return self.__class__(id=self.id, name=name)


def is_group_field(field):
    """Determines whether the given field is a group field.

    Args:
        field: a :class:`fiftyone.core.fields.Field`

    Returns:
        True/False
    """
    return isinstance(field, fof.EmbeddedDocumentField) and issubclass(
        field.document_type, Group
    )



"""
FiftyOne v0.16.3 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    pass


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    dataset_dict.pop("tags", None)

    db.datasets.replace_one(match_d, dataset_dict)



"""
DICOM utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import os
import warnings

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
import fiftyone.utils.data as foud

fou.ensure_package("pydicom")
import pydicom
from pydicom.fileset import FileInstance, FileSet


logger = logging.getLogger(__name__)


class DICOMSampleParser(foud.LabeledImageSampleParser):
    """Parser for labeled image samples stored in
    `DICOM format <https://en.wikipedia.org/wiki/DICOM>`_.

    Args:
        keywords (None): an optional keyword or list of keywords from
            :meth:`pydicom:pydicom.dataset.Dataset.dir` to load from the DICOM
            files. By default, all available fields are loaded
        parsers (None): an optional dict mapping keywords to functions that
            parse the values generated by
            :meth:`pydicom:pydicom.dataset.Dataset.get` for certain fields. By
            default, all fields are converted to primitive values, if possible
    """

    def __init__(self, keywords=None, parsers=None):
        if etau.is_str(keywords):
            keywords = [keywords]

        super().__init__()

        self.keywords = keywords
        self.parsers = parsers
        self._ds = None

    @property
    def label_cls(self):
        return None

    @property
    def has_image_path(self):
        return False

    @property
    def has_image_metadata(self):
        return False

    def get_image(self):
        self._ensure_ds()
        return _get_image(self._ds)

    def get_label(self):
        self._ensure_ds()

        if self.keywords is not None:
            fields = self.keywords
            blacklist = set()
        else:
            fields = self._ds.dir()
            blacklist = {"PixelData", "DataSetTrailingPadding"}

        label = {}

        for kw in fields:
            if kw in blacklist:
                continue

            value = self._ds.get(kw, default=None)

            if self.parsers is not None and kw in self.parsers:
                parser = self.parsers[kw]
                value = parser(value)

            if value is None:
                continue

            _value, success = _to_python(value)

            if not success:
                msg = "Ignoring field '%s' with unsupported type %s" % (
                    kw,
                    type(value),
                )
                warnings.warn(msg)
                continue

            label[kw] = _value

        return label

    def clear_sample(self):
        super().clear_sample()
        self._ds = None

    def _ensure_ds(self):
        if self._ds is None:
            if isinstance(self.current_sample, FileInstance):
                self._ds = self.current_sample.load()
            else:
                with open(self.current_sample, "rb") as f:
                    self._ds = pydicom.dcmread(f)


class DICOMDatasetImporter(
    foud.LabeledImageDatasetImporter, foud.ImportPathsMixin
):
    """Importer for DICOM datasets datasets stored on disk.

    See :ref:`this page <DICOMDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``dicom_path``
            must be provided
        dicom_path (None): an optional parameter that enables explicit control
            over the location of the DICOM files. Can be any of the following:

            -   a glob pattern like ``"*.dcm"`` specifying the location of the
                DICOM files in ``dataset_dir``
            -   the name of a DICOMDIR file in ``dataset_dir``
            -   an absolute glob pattern of DICOM files or the absolute path to
                a DICOMDIR file. In this case, ``dataset_dir`` has no effect

            If None, the parameter will default to ``*.dcm``
        images_dir (None): the directory in which the images will be written.
            If not provided, the images will be unpacked into ``dataset_dir``
        keywords (None): an optional keyword or list of keywords from
            :meth:`pydicom:pydicom.dataset.Dataset.dir` to load from the DICOM
            files. By default, all available fields are loaded
        parsers (None): an optional dict mapping keywords to functions that
            parse the values generated by
            :meth:`pydicom:pydicom.dataset.Dataset.get` for certain fields. By
            default, all fields are converted to primitive values, if possible
        image_format (None): the image format to use to write the images to
            disk. By default, ``fiftyone.config.default_image_ext`` is used
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        dicom_path=None,
        images_dir=None,
        keywords=None,
        parsers=None,
        image_format=None,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        if dataset_dir is None and dicom_path is None:
            raise ValueError(
                "Either `dataset_dir` or `dicom_path` must be provided"
            )

        dicom_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=dicom_path,
            default="*.dcm",
        )

        if images_dir is None:
            images_dir = os.path.dirname(dicom_path)
            logger.warning(
                "No `images_dir` provided. Images will be unpacked to '%s'",
                images_dir,
            )

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.dicom_path = dicom_path
        self.images_dir = images_dir
        self.keywords = keywords
        self.parsers = parsers
        self.image_format = image_format

        self._sample_parser = DICOMSampleParser(
            keywords=keywords, parsers=parsers
        )
        self._dataset_ingestor = None
        self._iter_dataset_ingestor = None
        self._num_samples = None

    def __len__(self):
        return self._num_samples

    def __iter__(self):
        self._iter_dataset_ingestor = iter(self._dataset_ingestor)
        return self

    def __next__(self):
        return next(self._iter_dataset_ingestor)

    @property
    def label_cls(self):
        return None

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_image_metadata(self):
        return self._sample_parser.has_image_metadata

    def setup(self):
        if os.path.isfile(self.dicom_path):
            if not os.path.splitext(self.dicom_path)[1]:
                # DICOMDIR file
                with open(self.dicom_path, "rb") as f:
                    ds = pydicom.dcmread(f)

                samples = list(FileSet(ds))
            else:
                # Single DICOM file
                samples = [self.dicom_path]
        else:
            # Glob pattern of DICOM files
            samples = etau.get_glob_matches(self.dicom_path)

        samples = self._preprocess_list(samples)
        self._num_samples = len(samples)

        self._dataset_ingestor = foud.LabeledImageDatasetIngestor(
            self.images_dir,
            samples,
            self._sample_parser,
            image_format=self.image_format,
        )
        self._dataset_ingestor.setup()

    def close(self, *args):
        self._dataset_ingestor.close(*args)


def _get_image(ds):
    # @todo allow non 8-bit images here?

    img = ds.pixel_array

    low = ds.get("SmallestImagePixelValue", 0)
    high = ds.get("LargestImagePixelValue", None)
    if high is None:
        high = img.max()

    return ((255 / max(high - low, 1)) * (img - low)).astype(np.uint8)


def _to_python(value):
    vtype = type(value)

    if issubclass(vtype, _LIST_TYPES):
        ctype = value.type_constructor

        if ctype in _PRIMITIVE_TYPES:
            return [ctype(v) for v in value], True

        ctype = _SCALAR_FIELD_TYPES_MAP.get(ctype, None)
        if ctype is not None:
            return [ctype(v) for v in value], True

        return None, False

    if vtype in _PRIMITIVE_TYPES:
        return value, True

    vtype = _SCALAR_FIELD_TYPES_MAP.get(vtype, None)
    if vtype is not None:
        return vtype(value), True

    return None, False


_PRIMITIVE_TYPES = (int, float, str, list)
_LIST_TYPES = (pydicom.multival.MultiValue,)

_SCALAR_FIELD_TYPES_MAP = {
    # pydicom.valuerep.DA: datetime.date,
    # pydicom.valuerep.DT: datetime.datetime,
    pydicom.valuerep.DSfloat: float,
    pydicom.valuerep.DSdecimal: float,
    pydicom.valuerep.IS: int,
    pydicom.valuerep.PersonName: str,
    pydicom.valuerep.PersonNameUnicode: str,
    pydicom.uid.UID: str,
}



"""
Dataset samples.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os

from fiftyone.core.document import Document, DocumentView
import fiftyone.core.frame as fofr
import fiftyone.core.frame_utils as fofu
import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
import fiftyone.core.media as fomm
import fiftyone.core.odm as foo
import fiftyone.core.utils as fou
from fiftyone.core.singletons import SampleSingleton


def get_default_sample_fields(include_private=False, use_db_fields=False):
    """Returns the default fields present on all samples.

    Args:
        include_private (False): whether to include fields starting with ``_``
        use_db_fields (False): whether to return database fields rather than
            user-facing fields, when applicable

    Returns:
        a tuple of field names
    """
    return foo.get_default_fields(
        foo.DatasetSampleDocument,
        include_private=include_private,
        use_db_fields=use_db_fields,
    )


class _SampleMixin(object):
    def __getattr__(self, name):
        if name == "frames" and self.media_type == fomm.VIDEO:
            return self._frames

        return super().__getattr__(name)

    def __setattr__(self, name, value):
        if name == "frames" and self.media_type == fomm.VIDEO:
            self.set_field("frames", value)
            return

        self._secure_media(name, value)
        super().__setattr__(name, value)

    def __getitem__(self, field_name):
        if self.media_type == fomm.VIDEO and fofu.is_frame_number(field_name):
            return self.frames[field_name]

        return super().__getitem__(field_name)

    def __setitem__(self, field_name, value):
        if self.media_type == fomm.VIDEO and fofu.is_frame_number(field_name):
            self.frames[field_name] = value
            return

        self._secure_media(field_name, value)
        super().__setitem__(field_name, value)

    def __iter__(self):
        if self.media_type == fomm.VIDEO:
            return iter(self._frames)

        raise ValueError("Image samples are not iterable")

    @property
    def filename(self):
        """The basename of the media's filepath."""
        return os.path.basename(self.filepath)

    @property
    def media_type(self):
        """The media type of the sample."""
        return self._media_type

    def get_field(self, field_name):
        if field_name == "frames" and self.media_type == fomm.VIDEO:
            return self._frames

        return super().get_field(field_name)

    def set_field(
        self,
        field_name,
        value,
        create=True,
        validate=True,
        dynamic=False,
    ):
        if field_name == "frames" and self.media_type == fomm.VIDEO:
            self.frames.clear()
            self.frames.update(
                value,
                expand_schema=create,
                validate=validate,
                dynamic=dynamic,
            )

            return

        super().set_field(
            field_name,
            value,
            create=create,
            validate=validate,
            dynamic=dynamic,
        )

    def clear_field(self, field_name):
        if field_name == "frames" and self.media_type == fomm.VIDEO:
            self.frames.clear()
            return

        super().clear_field(field_name)

    def compute_metadata(self, overwrite=False, skip_failures=False):
        """Populates the ``metadata`` field of the sample.

        Args:
            overwrite (False): whether to overwrite existing metadata
            skip_failures (False): whether to gracefully continue without
                raising an error if metadata cannot be computed
        """
        fom.compute_sample_metadata(
            self, overwrite=overwrite, skip_failures=skip_failures
        )

    def add_labels(
        self,
        labels,
        label_field=None,
        confidence_thresh=None,
        expand_schema=True,
        validate=True,
        dynamic=False,
    ):
        """Adds the given labels to the sample.

        The provided ``labels`` can be any of the following:

        -   A :class:`fiftyone.core.labels.Label` instance, in which case the
            labels are directly saved in the specified ``label_field``

        -   A dict mapping keys to :class:`fiftyone.core.labels.Label`
            instances. In this case, the labels are added as follows::

                for key, value in labels.items():
                    sample[label_key(key)] = value

        -   A dict mapping frame numbers to :class:`fiftyone.core.labels.Label`
            instances. In this case, the provided labels are interpreted as
            frame-level labels that should be added as follows::

                sample.frames.merge(
                    {
                        frame_number: {label_field: label}
                        for frame_number, label in labels.items()
                    }
                )

        -   A dict mapping frame numbers to dicts mapping keys to
            :class:`fiftyone.core.labels.Label` instances. In this case, the
            provided labels are interpreted as frame-level labels that should
            be added as follows::

                sample.frames.merge(
                    {
                        frame_number: {
                            label_key(key): value
                            for key, value in frame_dict.items()
                        }
                        for frame_number, frame_dict in labels.items()
                    }
                )

        In the above, the ``label_key`` function maps label dict keys to field
        names, and is defined from ``label_field`` as follows::

            if isinstance(label_field, dict):
                label_key = lambda k: label_field.get(k, k)
            elif label_field is not None:
                label_key = lambda k: label_field + "_" + k
            else:
                label_key = lambda k: k

        Args:
            labels: a :class:`fiftyone.core.labels.Label` or dict of labels per
                the description above
            label_field (None): the sample field, prefix, or dict defining in
                which field(s) to save the labels
            confidence_thresh (None): an optional confidence threshold to apply
                to any applicable labels before saving them
            expand_schema (True): whether to dynamically add new fields
                encountered to the dataset schema. If False, an error is raised
                if any fields are not in the dataset schema
            validate (True): whether to validate values for existing fields
            dynamic (False): whether to declare dynamic attributes
        """
        if isinstance(label_field, dict):
            label_key = lambda k: label_field.get(k, k)
        elif label_field is not None:
            label_key = lambda k: label_field + "_" + k
        else:
            label_key = lambda k: k

        if confidence_thresh is not None:
            labels = _apply_confidence_thresh(labels, confidence_thresh)

        if _is_frames_dict(labels):
            if self.media_type != fomm.VIDEO:
                raise ValueError(
                    "Cannot add frame labels to non-video samples"
                )

            if isinstance(next(iter(labels.values())), dict):
                # Multiple frame-level fields
                self.frames.merge(
                    {
                        frame_number: {
                            label_key(fname): flabel
                            for fname, flabel in frame_dict.items()
                        }
                        for frame_number, frame_dict in labels.items()
                    },
                    expand_schema=expand_schema,
                    validate=validate,
                    dynamic=dynamic,
                )
            elif label_field is None:
                raise ValueError(
                    "A `label_field` must be provided in order to add labels "
                    "to a single frame-level field"
                )
            else:
                # Single frame-level field
                self.frames.merge(
                    {
                        frame_number: {label_field: label}
                        for frame_number, label in labels.items()
                    },
                    expand_schema=expand_schema,
                    validate=validate,
                    dynamic=dynamic,
                )

        elif isinstance(labels, dict):
            # Multiple sample-level fields
            self.update_fields(
                {label_key(k): v for k, v in labels.items()},
                expand_schema=expand_schema,
                validate=validate,
                dynamic=dynamic,
            )
        elif labels is not None:
            if label_field is None:
                raise ValueError(
                    "A `label_field` must be provided in order to add labels "
                    "to a single sample field"
                )

            # Single sample-level field
            self.set_field(
                label_field,
                labels,
                create=expand_schema,
                validate=validate,
                dynamic=dynamic,
            )

        if self._in_db:
            self.save()

    def merge(
        self,
        sample,
        fields=None,
        omit_fields=None,
        merge_lists=True,
        overwrite=True,
        expand_schema=True,
        validate=True,
        dynamic=False,
    ):
        """Merges the fields of the given sample into this sample.

        The behavior of this method is highly customizable. By default, all
        top-level fields from the provided sample are merged in, overwriting
        any existing values for those fields, with the exception of list fields
        (e.g., ``tags``) and label list fields (e.g.,
        :class:`fiftyone.core.labels.Detections` fields), in which case the
        elements of the lists themselves are merged. In the case of label list
        fields, labels with the same ``id`` in both samples are updated rather
        than duplicated.

        To avoid confusion between missing fields and fields whose value is
        ``None``, ``None``-valued fields are always treated as missing while
        merging.

        This method can be configured in numerous ways, including:

        -   Whether new fields can be added to the dataset schema
        -   Whether list fields should be treated as ordinary fields and merged
            as a whole rather than merging their elements
        -   Whether to merge only specific fields, or all but certain fields
        -   Mapping input sample fields to different field names of this sample

        Args:
            sample: a :class:`fiftyone.core.sample.Sample`
            fields (None): an optional field or iterable of fields to which to
                restrict the merge. May contain frame fields for video samples.
                This can also be a dict mapping field names of the input sample
                to field names of this sample
            omit_fields (None): an optional field or iterable of fields to
                exclude from the merge. May contain frame fields for video
                samples
            merge_lists (True): whether to merge the elements of list fields
                (e.g., ``tags``) and label list fields (e.g.,
                :class:`fiftyone.core.labels.Detections` fields) rather than
                merging the entire top-level field like other field types.
                For label lists fields, existing
                :class:`fiftyone.core.label.Label` elements are either replaced
                (when ``overwrite`` is True) or kept (when ``overwrite`` is
                False) when their ``id`` matches a label from the provided
                sample
            overwrite (True): whether to overwrite (True) or skip (False)
                existing fields and label elements
            expand_schema (True): whether to dynamically add new fields
                encountered to the dataset schema. If False, an error is raised
                if any fields are not in the dataset schema
            validate (True): whether to validate values for existing fields
            dynamic (False): whether to declare dynamic embedded document
                fields
        """
        if sample.media_type != self.media_type:
            raise ValueError(
                "Cannot merge sample with media type '%s' into sample with "
                "media type '%s'" % (sample.media_type, self.media_type)
            )

        if self.media_type == fomm.VIDEO:
            (
                fields,
                frame_fields,
                omit_fields,
                omit_frame_fields,
            ) = self._parse_fields_video(
                fields=fields, omit_fields=omit_fields
            )

        super().merge(
            sample,
            fields=fields,
            omit_fields=omit_fields,
            merge_lists=merge_lists,
            overwrite=overwrite,
            expand_schema=expand_schema,
            validate=validate,
            dynamic=dynamic,
        )

        if self.media_type == fomm.VIDEO:
            self.frames.merge(
                sample.frames,
                fields=frame_fields,
                omit_fields=omit_frame_fields,
                merge_lists=merge_lists,
                overwrite=overwrite,
                expand_schema=expand_schema,
                validate=validate,
                dynamic=dynamic,
            )

    def copy(self, fields=None, omit_fields=None):
        """Returns a deep copy of the sample that has not been added to the
        database.

        Args:
            fields (None): an optional field or iterable of fields to which to
                restrict the copy. This can also be a dict mapping existing
                field names to new field names
            omit_fields (None): an optional field or iterable of fields to
                exclude from the copy

        Returns:
            a :class:`Sample`
        """
        if self.media_type == fomm.VIDEO:
            (
                fields,
                frame_fields,
                omit_fields,
                omit_frame_fields,
            ) = self._parse_fields_video(
                fields=fields, omit_fields=omit_fields
            )

        sample = super().copy(fields=fields, omit_fields=omit_fields)

        if self.media_type == fomm.VIDEO:
            sample.frames.update(
                {
                    frame_number: frame.copy(
                        fields=frame_fields, omit_fields=omit_frame_fields
                    )
                    for frame_number, frame in self.frames.items()
                }
            )

        return sample

    def to_dict(self, include_frames=False, include_private=False):
        """Serializes the sample to a JSON dictionary.

        Args:
            include_frames (False): whether to include the frame labels for
                video samples
            include_private (False): whether to include private fields

        Returns:
            a JSON dict
        """
        d = super().to_dict(include_private=include_private)

        if self.media_type == fomm.VIDEO:
            if include_frames:
                d["frames"] = self.frames._to_frames_dict(
                    include_private=include_private
                )
            else:
                d.pop("frames", None)

        return d

    def _secure_media(self, field_name, value):
        if field_name != "filepath":
            return

        new_media_type = fomm.get_media_type(value)
        if self.media_type != new_media_type:
            raise fomm.MediaTypeError(
                "A sample's 'filepath' can be changed, but its media type "
                "cannot; current '%s', new '%s'"
                % (self.media_type, new_media_type)
            )

    def _parse_fields_video(self, fields=None, omit_fields=None):
        if fields is not None:
            fields, frame_fields = fou.split_frame_fields(fields)
        else:
            frame_fields = None

        if omit_fields is not None:
            omit_fields, omit_frame_fields = fou.split_frame_fields(
                omit_fields
            )
        else:
            omit_frame_fields = None

        return fields, frame_fields, omit_fields, omit_frame_fields


class Sample(_SampleMixin, Document, metaclass=SampleSingleton):
    """A sample in a :class:`fiftyone.core.dataset.Dataset`.

    Samples store all information associated with a particular piece of data in
    a dataset, including basic metadata about the data, one or more sets of
    labels (ground truth, user-provided, or FiftyOne-generated), and additional
    features associated with subsets of the data and/or label sets.

    .. note::

        :class:`Sample` instances that are **in datasets** are singletons,
        i.e.,  ``dataset[sample_id]`` will always return the same
        :class:`Sample` instance.

    Args:
        filepath: the path to the data on disk. The path is converted to an
            absolute path (if necessary) via
            :func:`fiftyone.core.utils.normalize_path`
        tags (None): a list of tags for the sample
        metadata (None): a :class:`fiftyone.core.metadata.Metadata` instance
        **kwargs: additional fields to dynamically set on the sample
    """

    _NO_DATASET_DOC_CLS = foo.NoDatasetSampleDocument

    def __init__(self, filepath, tags=None, metadata=None, **kwargs):
        super().__init__(
            filepath=filepath, tags=tags, metadata=metadata, **kwargs
        )

        if self.media_type == fomm.VIDEO:
            self._frames = fofr.Frames(self)
        else:
            self._frames = None

    def __repr__(self):
        kwargs = {}
        if self.media_type == fomm.VIDEO:
            kwargs["frames"] = self._frames

        return self._doc.fancy_repr(
            class_name=self.__class__.__name__, **kwargs
        )

    def _reload_backing_doc(self):
        if not self._in_db:
            return

        d = self._dataset._sample_collection.find_one({"_id": self._id})
        self._doc = self._dataset._sample_dict_to_doc(d)

    def reload(self, hard=False):
        """Reloads the sample from the database.

        Args:
            hard (False): whether to reload the sample's schema in addition to
                its field values. This is necessary if new fields may have been
                added to the dataset schema
        """
        if self.media_type == fomm.VIDEO:
            self.frames.reload(hard=hard)

        super().reload(hard=hard)

    def save(self):
        """Saves the sample to the database."""
        super().save()

    def _save(self, deferred=False):
        if not self._in_db:
            raise ValueError(
                "Cannot save a sample that has not been added to a dataset"
            )

        if self.media_type == fomm.VIDEO:
            frame_ops = self.frames._save(deferred=deferred)
        else:
            frame_ops = []

        sample_op = super()._save(deferred=deferred)

        return sample_op, frame_ops

    @classmethod
    def from_frame(cls, frame, filepath=None):
        """Creates a sample from the given frame.

        Args:
            frame: a :class:`fiftyone.core.frame.Frame`
            filepath (None): the path to the corresponding image frame on disk,
                if not available

        Returns:
            a :class:`Sample`
        """
        kwargs = {k: v for k, v in frame.iter_fields()}
        if filepath is not None:
            kwargs["filepath"] = filepath

        return cls(**kwargs)

    @classmethod
    def from_doc(cls, doc, dataset=None):
        """Creates a sample backed by the given document.

        Args:
            doc: a :class:`fiftyone.core.odm.sample.DatasetSampleDocument` or
                :class:`fiftyone.core.odm.sample.NoDatasetSampleDocument`
            dataset (None): the :class:`fiftyone.core.dataset.Dataset` that
                the sample belongs to

        Returns:
            a :class:`Sample`
        """
        sample = super().from_doc(doc, dataset=dataset)

        if sample.media_type == fomm.VIDEO:
            sample._frames = fofr.Frames(sample)

        return sample

    @classmethod
    def from_dict(cls, d):
        """Loads the sample from a JSON dictionary.

        The returned sample will not belong to a dataset.

        Returns:
            a :class:`Sample`
        """
        media_type = d.pop("_media_type", None)
        if media_type is None:
            media_type = fomm.get_media_type(d.get("filepath", ""))

        if media_type == fomm.VIDEO:
            frames = d.pop("frames", {})

        sample = super().from_dict(d)

        if sample.media_type == fomm.VIDEO:
            for fn, fd in frames.items():
                sample.frames[int(fn)] = fofr.Frame.from_dict(fd)

        return sample


class SampleView(_SampleMixin, DocumentView):
    """A view into a :class:`Sample` in a dataset.

    Like :class:`Sample` instances, the fields of a :class:`SampleView`
    instance can be modified, new fields can be created, and any changes can be
    saved to the database.

    :class:`SampleView` instances differ from :class:`Sample` instances in the
    following ways:

    -   A sample view may contain only a subset of the fields of its source
        sample, either by selecting and/or excluding specific fields
    -   A sample view may contain array fields or embedded array fields that
        have been filtered, thus containing only a subset of the array elements
        from the source sample
    -   Excluded fields of a sample view may not be accessed or modified

    .. note::

        Sample views should never be created manually; they are generated when
        accessing the samples in a :class:`fiftyone.core.view.DatasetView`.

    Args:
        doc: a :class:`fiftyone.core.odm.mixins.DatasetSampleDocument`
        view: the :class:`fiftyone.core.view.DatasetView` that the sample
            belongs to
        selected_fields (None): a set of field names that this sample view is
            restricted to, if any
        excluded_fields (None): a set of field names that are excluded from
            this sample view, if any
        filtered_fields (None): a set of field names of list fields that are
            filtered in this sample view, if any
    """

    _DOCUMENT_CLS = Sample

    def __init__(
        self,
        doc,
        view,
        selected_fields=None,
        excluded_fields=None,
        filtered_fields=None,
    ):
        super().__init__(
            doc,
            view,
            selected_fields=selected_fields,
            excluded_fields=excluded_fields,
            filtered_fields=filtered_fields,
        )

        if self.media_type == fomm.VIDEO:
            self._frames = fofr.FramesView(self)
        else:
            self._frames = None

    def __repr__(self):
        if self._selected_fields is not None:
            select_fields = ("media_type",) + tuple(self._selected_fields)
        else:
            select_fields = None

        kwargs = {}
        if self.media_type == fomm.VIDEO:
            kwargs["frames"] = self._frames

        return self._doc.fancy_repr(
            class_name=self.__class__.__name__,
            select_fields=select_fields,
            exclude_fields=self._excluded_fields,
            **kwargs,
        )

    def to_dict(self, include_frames=False, include_private=False):
        """Serializes the sample view to a JSON dictionary.

        Args:
            include_frames (False): whether to include the frame labels for
                video samples
            include_private (False): whether to include private fields

        Returns:
            a JSON dict
        """
        d = super().to_dict(
            include_frames=include_frames, include_private=include_private
        )

        if self.selected_field_names or self.excluded_field_names:
            field_names = set(
                self._get_field_names(
                    include_private=include_private,
                    use_db_fields=True,
                )
            )

            if include_frames and self.media_type == fomm.VIDEO:
                field_names.add("frames")

            d = {k: v for k, v in d.items() if k in field_names}

        return d

    def save(self):
        """Saves the sample view to the database.

        .. warning::

            This will permanently delete any omitted or filtered contents from
            the source dataset.
        """
        super().save()

    def _save(self, deferred=False):
        if self.media_type == fomm.VIDEO:
            frame_ops = self.frames._save(deferred=deferred)
        else:
            frame_ops = []

        sample_op = super()._save(deferred=deferred)

        return sample_op, frame_ops


def _apply_confidence_thresh(label, confidence_thresh):
    if _is_frames_dict(label):
        label = {
            frame_number: _apply_confidence_thresh(
                frame_dict, confidence_thresh
            )
            for frame_number, frame_dict in label.items()
        }
    elif isinstance(label, dict):
        label = {
            k: _apply_confidence_thresh(v, confidence_thresh)
            for k, v in label.items()
        }
    elif isinstance(label, fol._LABEL_LIST_FIELDS):
        labels = [
            l
            for l in getattr(label, label._LABEL_LIST_FIELD)
            if l.confidence is not None and l.confidence >= confidence_thresh
        ]
        setattr(label, label._LABEL_LIST_FIELD, labels)
    elif hasattr(label, "confidence"):
        if label.confidence is None or label.confidence < confidence_thresh:
            label = None

    return label


def _is_frames_dict(label):
    return (
        label
        and isinstance(label, dict)
        and fofu.is_frame_number(next(iter(label.keys())))
    )



"""
Utilities for working with the `ImageNet dataset <http://www.image-net.org>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os


_TRAIN_IMAGES_TAR = "ILSVRC2012_img_train.tar"
_VAL_IMAGES_DIR = "ILSVRC2012_img_val.tar"
_DEVKIT_TAR = "ILSVRC2012_devkit_t12.tar.gz"


def ensure_imagenet_manual_download(source_dir, split, devkit=False):
    """Ensures that the ImageNet archive(s) for the requested split have been
    manually downloaded to the required locations.

    Args:
        source_dir: the dataset directory
        split: the split of interest. Supported values are
            ``("train", "validation")``
        devkit (False): whether to ensure that the devkit archive is present

    Raises:
        OSError: if the required files are not present
    """
    if split == "train":
        archive_name = _TRAIN_IMAGES_TAR
    elif split == "validation":
        archive_name = _VAL_IMAGES_DIR
    else:
        raise ValueError(
            "Unsupported split '%s'; Supported values are "
            "('train', 'validation')"
        )

    _ensure_archive(archive_name, source_dir)

    if devkit:
        devkit_name = _DEVKIT_TAR
        _ensure_archive(devkit_name, source_dir)


def _ensure_archive(archive_name, source_dir):
    if source_dir is None:
        _raise_imagenet_error(
            "You must provide a `source_dir` in order to load the ImageNet "
            "dataset."
        )

    archive_path = os.path.join(source_dir, archive_name)
    if not os.path.isfile(archive_path):
        _raise_imagenet_error(
            "Archive '%s' not found in directory '%s'."
            % (archive_name, source_dir)
        )


def _raise_imagenet_error(msg):
    raise OSError(
        "\n\n"
        + msg
        + "\n\n"
        + "You must download the source files for the ImageNet dataset "
        "manually."
        + "\n\n"
        + "Run `fiftyone zoo datasets info imagenet-2012` for more information"
    )



"""
Defines the shared state between the FiftyOne App and backend.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from bson import json_util
from dataclasses import asdict
import json
import logging
import typing as t

import strawberry as gql

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.dataset as fod
import fiftyone.core.media as fom
import fiftyone.core.utils as fou
import fiftyone.core.view as fov
from fiftyone.server.scalars import JSON


logger = logging.getLogger(__name__)


class StateDescription(etas.Serializable):
    """Class that describes the shared state between the FiftyOne App and
    a corresponding :class:`fiftyone.core.session.Session`.

    Args:
        config (None): an optional :class:`fiftyone.core.config.AppConfig`
        dataset (None): the current :class:`fiftyone.core.dataset.Dataset`
        selected (None): the list of currently selected samples
        selected_labels (None): the list of currently selected labels
        view (None): the current :class:`fiftyone.core.view.DatasetView`
    """

    def __init__(
        self,
        config=None,
        dataset=None,
        selected=None,
        selected_labels=None,
        view=None,
    ):
        self.config = config or fo.app_config.copy()
        self.dataset = dataset
        self.selected = selected or []
        self.selected_labels = selected_labels or []
        self.view = view

    def serialize(self, reflective=False):
        with fou.disable_progress_bars():
            d = super().serialize(reflective=reflective)

            if self.dataset is not None:
                d["dataset"] = self.dataset.name
                collection = self.dataset
                if self.view is not None:
                    collection = self.view
                    d["view"] = json.loads(
                        json_util.dumps(self.view._serialize())
                    )
                    d["view_cls"] = etau.get_class_name(self.view)

                d["sample_fields"] = serialize_fields(
                    collection.get_field_schema(flat=True), dicts=True
                )
                d["frame_fields"] = serialize_fields(
                    collection.get_frame_field_schema(flat=True), dicts=True
                )

                view = self.view if self.view is not None else self.dataset
                if view.media_type == fom.GROUP:
                    d["group_slice"] = view.group_slice

            d["config"]["timezone"] = fo.config.timezone

            if self.config.colorscale:
                d["colorscale"] = self.config.get_colormap()

            return d

    def attributes(self):
        return list(
            filter(
                lambda a: a not in {"dataset", "view"}, super().attributes()
            )
        )

    @classmethod
    def from_dict(cls, d, with_config=None):
        """Constructs a :class:`StateDescription` from a JSON dictionary.

        Args:
            d: a JSON dictionary
            with_config (None): an existing
                :class:`fiftyone.core.config.AppConfig` to attach and apply
                settings to

        Returns:
            :class:`StateDescription`
        """
        dataset = d.get("dataset", None)
        if dataset is not None:
            dataset = fod.load_dataset(dataset)

        stages = d.get("view", None)
        if dataset is not None and stages:
            view = fov.DatasetView._build(dataset, stages)
        else:
            view = None

        group_slice = d.get("group_slice", None)
        if group_slice:
            if dataset is not None:
                dataset.group_slice = group_slice

            if view is not None:
                view.group_slice = group_slice

        config = with_config or fo.app_config.copy()
        for field, value in d.get("config", {}).items():
            setattr(config, field, value)

        fo.config.timezone = d.get("config", {}).get("timezone", None)

        return cls(
            config=config,
            dataset=dataset,
            selected=d.get("selected", []),
            selected_labels=d.get("selected_labels", []),
            view=view,
        )


@gql.type
class SampleField:
    ftype: str
    path: str
    subfield: t.Optional[str]
    embedded_doc_type: t.Optional[str]
    db_field: t.Optional[str]
    description: t.Optional[str]
    info: t.Optional[JSON]


def serialize_fields(schema: t.Dict, dicts=False) -> t.List[SampleField]:
    data = []

    if schema:
        for path, field in schema.items():
            if isinstance(field, fo.EmbeddedDocumentField):
                embedded_doc_type = etau.get_class_name(field.document_type)
            elif (
                isinstance(field, fo.ListField)
                and field.field
                and isinstance(field.field, fo.EmbeddedDocumentField)
            ):
                embedded_doc_type = etau.get_class_name(
                    field.field.document_type
                )
            else:
                embedded_doc_type = None

            if (
                isinstance(field, (fo.DictField, fo.ListField))
                and field.field is not None
            ):
                subfield = etau.get_class_name(field.field)
            else:
                subfield = None

            if field.info is not None:
                # Converts mongoengine types to primitives
                info = json.loads(json.dumps(field.info))
            else:
                info = None

            data.append(
                SampleField(
                    path=path,
                    db_field=field.db_field,
                    ftype=etau.get_class_name(field),
                    embedded_doc_type=embedded_doc_type,
                    subfield=subfield,
                    description=field.description,
                    info=info,
                )
            )

    if dicts:
        return [asdict(f) for f in data]

    return data
